@@ -82,13 +82,14 @@ module Ast : sig
   end
 
   module Term : sig
+    type impl_kind = L2R | L2RBang | R2L
     type t_ =
-      | Impl of bool * t * t
+      | Impl of impl_kind * Loc.t * t * t
       | Const of Scope.t * Name.t
       | Discard
-      | Var of Name.t * t list (** unification variable *)
-      | App of Scope.t * Name.t * t * t list
-      | Lam of (Name.t * Scope.language) option * Type.t option * t
+      | Var of Name.t * Loc.t * t list (** unification variable *)
+      | App of Scope.t * Name.t * Loc.t * t * t list
+      | Lam of (Name.t * Loc.t * Scope.language) option * Type.t option * t
       | Opaque of Opaque.t
       | Cast of t * Type.t
     and t = { it : t_; loc : Loc.t; }
@@ -99,23 +100,23 @@ module Ast : sig
     type constant = Name.constant
     val mkGlobal : loc:Loc.t -> constant -> t
     val mkBound : loc:Loc.t -> language:Scope.language -> Name.t -> t
-    val mkAppGlobal : loc:Loc.t -> constant -> t -> t list -> t
-    val mkAppBound : loc:Loc.t ->  language:Scope.language -> Name.t -> t -> t list -> t
-    val mkVar : loc:Loc.t -> Name.t -> t list -> t
+    val mkAppGlobal : loc:Loc.t -> hdloc:Loc.t -> constant -> t -> t list -> t
+    val mkAppBound : loc:Loc.t -> hdloc:Loc.t -> language:Scope.language -> Name.t -> t -> t list -> t
+    val mkVar : loc:Loc.t -> hdloc:Loc.t -> Name.t -> t list -> t
     val mkOpaque : loc:Loc.t -> Opaque.t -> t
     val mkCast : loc:Loc.t -> t -> Type.t -> t
-    val mkLam : loc:Loc.t -> (Name.t *  Scope.language) option -> ?ty:Type.t -> t -> t
+    val mkLam : loc:Loc.t -> (Name.t * Loc.t * Scope.language) option -> ?ty:Type.t -> t -> t
     val mkDiscard : loc:Loc.t -> t
 
     (** Handy constructors to build goals *)
-    val mkImplication : loc:Loc.t -> t -> t -> t
-    val mkPi : loc:Loc.t -> Name.t -> ?ty:Type.t -> t -> t
-    val mkConj : loc:Loc.t -> t list -> t
-    val mkEq : loc:Loc.t -> t -> t -> t
+    val mkImplication : loc:Loc.t -> hdloc:Loc.t -> t -> t -> t
+    val mkPi : loc:Loc.t -> hdloc:Loc.t -> Name.t -> nloc:Loc.t -> ?ty:Type.t -> t -> t
+    val mkConj : loc:Loc.t -> hdloc:Loc.t -> t list -> t
+    val mkEq : loc:Loc.t -> hdloc:Loc.t -> t -> t -> t
     val mkNil : loc:Loc.t -> t
     (** if omitted, the loc is the merge of the hd and tl locs, as if
         one wrote (hd :: tl), but not as if one wrote [hd|tl] *)
-    val mkCons : ?loc:Loc.t -> t -> t -> t
+    val mkCons : ?loc:Loc.t -> hdloc:Loc.t -> t -> t -> t
 
     val list_to_lp_list : loc:Loc.t -> t list -> t
     val ne_list_to_lp_list : t list -> t
@@ -197,7 +198,8 @@ module Setup : sig
   val trace : string list -> string list
 
   (** Override default runtime error functions (they call exit) *)
-  val set_warn : (?loc:Ast.Loc.t -> string -> unit) -> unit
+  type warning_id = LinearVariable | UndeclaredGlobal | FlexClause | ImplicationPrecedence
+  val set_warn : (?loc:Ast.Loc.t -> id:warning_id -> string -> unit) -> unit
   val set_error : (?loc:Ast.Loc.t -> string -> 'a) -> unit
   val set_anomaly : (?loc:Ast.Loc.t -> string -> 'a) -> unit
   val set_type_error : (?loc:Ast.Loc.t -> string -> 'a) -> unit
@@ -252,6 +254,8 @@ module Compile : sig
     print_units : bool;
     (* keep track of the amount of time spent type checking, default false *)
     time_typechecking : bool;
+    (* skip determinacy analysis *)
+    skip_det_checking: bool;
   }
   val default_flags : flags
   val to_setup_flags : flags -> Setup.flags
@@ -305,6 +309,21 @@ module Compile : sig
   val optimize : query -> executable
   
   val total_type_checking_time : query -> float
+  val total_det_checking_time : query -> float
+
+  module IntervalTree : sig
+    type 'a t
+    val find : Ast.Loc.t -> 'a t -> (Ast.Loc.t * 'a) list
+    val pp : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
+  end
+
+  type type_
+  val pp_type_ : Format.formatter -> type_ -> unit
+
+  type info = { defined : Ast.Loc.t option; type_ : type_ option }
+  val pp_info : Format.formatter -> info -> unit
+
+  val hover : compilation_unit -> info IntervalTree.t
 end
 
 module Data : sig
@@ -448,6 +467,7 @@ module ContextualConversion : sig
 
   (* cast *)
   val (!<) : ('a,unit,unit) t -> 'a Conversion.t
+  val (!<<) : (('a,unit,unit) t -> ('b,unit,unit) t) -> ('a Conversion.t -> 'b Conversion.t)
 
   (* morphisms *)
   val (!>)   : 'a Conversion.t -> ('a,'hyps,'constraints) t
@@ -583,14 +603,23 @@ module AlgebraicData : sig
         ('match_stateful_t,'match_t,'t) match_t
       -> ('t,'h,'c) constructor
 
-  type ('t,'h,'c) declaration = {
+  type ('t,'h,'c) base_declaration = {
     ty : Conversion.ty_ast;
     doc : doc;
     pp : Format.formatter -> 't -> unit;
     constructors : ('t,'h,'c) constructor list;
   }
 
-  val declare : ('t,'h,'c) declaration -> ('t,'h,'c) ContextualConversion.t
+  type ('t,'h,'c) declaration =
+  | Decl : ('t,'h,'c) base_declaration -> ('t,'h,'c) declaration
+  | Param : ('t Conversion.t -> ('t1,'h,'c) declaration) -> ('t1,'h,'c) declaration
+  | ParamC : (('t,'h,'c) ContextualConversion.t -> ('t1,'h,'c) declaration) -> ('t1,'h,'c) declaration
+
+  type allocation
+
+  val allocate_constructors : ('t,'h,'c) declaration -> allocation
+  val declare_allocated : allocation -> ('t,'h,'c) declaration -> ('t,'h,'c) ContextualConversion.t
+  val declare : ('t,'h,'c) base_declaration -> ('t,'h,'c) ContextualConversion.t
 
 end
 
@@ -862,7 +891,6 @@ module BuiltIn : sig
 
 end
 
-
 (* ************************************************************************* *)
 (* ********************* Advanced Extension API **************************** *)
 (* ************************************************************************* *)
@@ -932,6 +960,10 @@ module FlexibleData : sig
     val equal : t -> t -> bool
     val hash : t -> int
     val fresh : unit -> Ast.Name.t
+
+    (** dbug print *)
+    val pp_raw :  Format.formatter -> t -> unit
+    val show_raw :  t -> string
   end
 
   module type Host = sig
@@ -1121,7 +1153,7 @@ module RawData : sig
       starts at 0 and grows for bound variables;
       global constants have negative values. *)
 
-  type builtin
+  type builtin = Cut | And | Impl | ImplBang | RImpl | Pi | Sigma | Eq | Match | Findall | Delay | Host of constant
   type term = Data.term
   type view = private
     (* Pure subterms *)
@@ -1138,8 +1170,8 @@ module RawData : sig
     | UnifVar of FlexibleData.Elpi.t * term list
 
   (** Terms must be inspected after dereferencing their head.
-      If the resulting term is UVar then its uvar_body is such that
-      get_assignment uvar_body = None *)
+      If the resulting term is UVar then its uvar is such that
+      get_assignment uvar = None *)
   val look : depth:int -> term -> view
 
   (* to reuse a term that was looked at *)
@@ -1178,35 +1210,22 @@ module RawData : sig
   val of_hyp : Data.hyp -> hyp
   val of_hyps : Data.hyp list -> hyps
 
+  type blockers = FlexibleData.Elpi.t list
   type suspended_goal = {
     context : hyps;
-    goal : int * term
+    goal : int * term;
+    blockers : blockers;
   }
   val constraints : Data.constraints -> suspended_goal list
   val no_constraints : Data.constraints
 
   module Constants : sig
 
-    val declare_global_symbol : string -> constant
+    val declare_global_symbol : ?variant:int -> string -> constant
 
     val show : constant -> string
 
-    val eqc    : builtin (* = *)
     val orc    : constant (* ; *)
-    val andc   : constant (* , *)
-    val rimplc : constant (* :- *)
-    val pic    : constant (* pi *)
-    val sigmac : constant (* sigma *)
-    val implc  : constant (* => *)
-    val cutc   : builtin (* ! *)
-
-    (* LambdaProlog built-in data types are just instances of CData.
-     * The typeabbrev machinery translates [int], [float] and [string]
-     * to [ctype "int"], [ctype "float"] and [ctype "string"]. *)
-    val ctypec : constant (* ctype *)
-
-    (* Marker for spilling function calls, as in [{ rev L }] *)
-    val spillc : constant
 
     module Map : Map.S with type key = constant
     module Set : Set.S with type elt = constant
@@ -1298,13 +1317,19 @@ module Utils : sig
   val type_error : ?loc:Ast.Loc.t ->string -> 'a
 
   (** A non fatal warning *)
-  val warn : ?loc:Ast.Loc.t ->string -> unit
+  val warn : ?loc:Ast.Loc.t -> id:Setup.warning_id -> string -> unit
 
   (** link between OCaml and LP lists. Note that [1,2|X] is not a valid
    * OCaml list! *)
   val list_to_lp_list : Data.term list -> Data.term
   val lp_list_to_list : depth:int -> Data.term -> Data.term list
 
+  (** no uvars, raises No_clause *)
+  val check_ground : depth:int -> Data.term -> unit 
+
+  (** comparison (need terms to be ground) *)
+  val cmp_term : depth:int -> Data.term -> Data.term -> int
+
   (** The body of an assignment, if any (LOW LEVEL).
    * Use [look] and forget about this API since the term you get
    * needs to be moved and/or reduced, and you have no API for this. *)
@@ -1336,6 +1361,12 @@ module Utils : sig
     val show : t -> string
   end
 
+  module type ShowKey = sig
+    type key
+    val pp_key : Format.formatter -> key -> unit
+    val show_key : key -> string
+  end
+
   module type Show1 = sig
     type 'a t
     val pp : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
