@@ -148,7 +148,6 @@ end = struct
       goal_attempts : goal_attempts GoalMap.t
     }
   
-let elaborated_steps : Elaborated_step.t StepMap.t ref = ref StepMap.empty
 
 (* elaboration *)
 let rec filter_map f = function
@@ -181,9 +180,6 @@ let decode_chr_store_entry = function
   | { payload = [ gid; gtext ]}, _ -> int_of_string gid, gtext
   | _ -> assert false
 
-let get s l =
-  try (fst @@ List.find (fun ({ name },_) -> name = s) l).payload
-  with Not_found -> assert false
 
 type decoded_step =
   [ `Focus of item * time
@@ -235,7 +231,7 @@ let int_of_string s =
 let decode_loc s =
   if Str.(string_match (regexp "File .(context step_id:\\([0-9]+\\))") s 0) then
     `Context (int_of_string (Str.matched_group 1 s))
-  else if Str.(string_match (regexp "File .\\([^,]+\\)., line \\([0-9]+\\), column \\([0-9]+\\), characters? \\([0-9]+\\)-?[0-9]*") s 0) then
+  else if Str.(string_match (regexp "File .\\([^,]+\\)., line \\([0-9]+\\), column \\([0-9]+\\), characters? -*\\([0-9]+\\)-*[0-9]*") s 0) then
     `File {
       filename = Str.matched_group 1 s;
       line = Str.matched_group 2 s |> int_of_string;
