@@ -53,8 +53,10 @@ let mkOutputMode = encode ~k:kOther ~data:2 ~arity:0
 let mkListTailVariable = encode ~k:kOther ~data:3 ~arity:0
 let mkListHead = encode ~k:kOther ~data:4 ~arity:0
 let mkListEnd = encode ~k:kOther ~data:5 ~arity:0 
-let mkPathEnd = encode ~k:kOther ~data:6 ~arity:0
-let mkListTailVariableUnif = encode ~k:kOther ~data:7 ~arity:0
+let mkListNil = encode ~k:kOther ~data:6 ~arity:0
+let mkPathEnd = encode ~k:kOther ~data:7 ~arity:0
+let mkListTailVariableUnif = encode ~k:kOther ~data:8 ~arity:0
+let mkUvarConstant = encode ~k:kVariable ~data:9 ~arity:0
 
 
 let isVariable x = x == mkVariable
@@ -62,10 +64,12 @@ let isAny x = x == mkAny
 let isInput x = x == mkInputMode
 let isOutput x = x == mkOutputMode
 let isListHead x = x == mkListHead
+let isListNil x = x == mkListNil
 let isListEnd x = x == mkListEnd
 let isListTailVariable x = x == mkListTailVariable
 let isListTailVariableUnif x = x == mkListTailVariableUnif
 let isPathEnd x = x == mkPathEnd
+let isUvarConstant x = x == mkUvarConstant
 
 type cell = int
 
@@ -84,8 +88,10 @@ let pp_cell fmt n =
        else if isListHead n then "ListHead"
        else if isListEnd n then "ListEnd"
        else if isPathEnd n then "PathEnd"
+       else if isListNil n then "ListNil"
        else if isListTailVariableUnif n then "ListTailVariableUnif"
        else if isAny n then "Other"
+       else if isUvarConstant n then "uvar"
        else failwith "Invalid path construct...")
   else if k == kPrimitive then Format.fprintf fmt "Primitive"
   else Format.fprintf fmt "%o" k
@@ -202,20 +208,21 @@ module Trie = struct
 
   let rec pp (ppelem : Format.formatter -> 'a -> unit) (fmt : Format.formatter)
       (Node { data; other; listTailVariable; map } : 'a t) : unit =
-    Format.fprintf fmt "@[<v>[values:{";
+    Format.fprintf fmt "@[<v>[@[<hov 2>values:{@ ";
     pplist ppelem "; " fmt data;
-    Format.fprintf fmt "}@ other:{";
+    Format.fprintf fmt "}@ @]@ @[<hov 2> other:{@ ";
     (match other with None -> () | Some m -> pp ppelem fmt m);
-    Format.fprintf fmt "}@ listTailVariable:{";
+    Format.fprintf fmt "}@ @]@ @[<hov 2> listTailVariable:{@ ";
     (match listTailVariable with None -> () | Some m -> pp ppelem fmt m);
-    Format.fprintf fmt "}@ key:{@[<hov 2>";
+    Format.fprintf fmt "}@ @]@ @[<hov 2> key:{@ @[<hov 2>";
     Ptmap.to_list map
     |> pplist
          (fun fmt (k, v) ->
-           pp_cell fmt k;
-           pp ppelem fmt v)
+          Format.fprintf fmt "@[<hov 2> %a@ %a@]"
+           pp_cell k
+           (pp ppelem) v)
          ";@ " fmt;
-    Format.fprintf fmt "@]}]@]"
+    Format.fprintf fmt "@]}@]]@]"
 
   let show (fmt : Format.formatter -> 'a -> unit) (n : 'a t) : string =
     let b = Buffer.create 22 in
@@ -254,10 +261,17 @@ let skip_listTailVariable ~pos path : int =
     else aux (i+1) (update_par_count pc (Path.get path i)) in
   aux (pos + 1) (update_par_count 1 (Path.get path pos))
 
-(* A discrimination tree is a record {t; max_size; max_depths } where
+(* A discrimination tree is a record {t; max_size; max_depths; max_list_length; user_upper_bound_depth } where
    - t is a Trie for instance retrival 
    - max_size is a int representing the max size of a path in the Trie
    - max_depths is a int list representing the max depth of the each indexed arguments
+   - max_list_length is the length of the maximum indexed list
+   - user_upper_bound_depth is the index defined by the user
+  
+  Invariants:
+    - length(max_depths) = length(user_upper_bound_depth)
+    - forall i in range (length max_depths):
+        assert (max_depths[i] <= user_upper_bound_depth[i])
 
   Note : 
   - max_size is used as an heuristic for the size of path when a new one is created
@@ -282,18 +296,19 @@ let skip_listTailVariable ~pos path : int =
     In the example it is no needed to index the goal path to depth 100, but rather considering
     the maximal depth of the first argument, which 4 << 100
   *)
-type 'a t = {t: 'a Trie.t; max_size : int;  max_depths : int array; max_list_length: int } 
+type 'a t = {t: 'a Trie.t; max_size : int;  max_depths : int array; max_list_length: int; user_upper_bound_depth:int list } 
 
 let pp pp_a fmt { t } : unit = Trie.pp (fun fmt data -> pp_a fmt data) fmt t
 let show pp_a { t } : string = Trie.show (fun fmt data -> pp_a fmt data) t
 
-let index { t; max_size; max_depths; max_list_length = mll } ~max_list_length path data =
-  let t, m = Trie.add path data t in
-  { t; max_size = max max_size m; max_depths; max_list_length = max max_list_length mll }
+let index { t; max_size; max_depths; max_list_length = mll; user_upper_bound_depth } ~max_list_length path data =
+  let t, (m : int) = Trie.add path data t in
+  { t; max_size = max max_size m; max_depths; max_list_length = max max_list_length mll; user_upper_bound_depth }
 
 let max_path { max_size } = max_size
 let max_depths { max_depths } = max_depths 
 let max_list_length { max_list_length } = max_list_length
+let user_upper_bound_depth t = t.user_upper_bound_depth
 
 (* the equivalent of skip, but on the index, thus the list of trees
     that are rooted just after the term represented by the tree root
@@ -343,14 +358,16 @@ let rec retrieve ~pos ~add_result mode path tree : unit =
     let sub_tries = skip_to_listEnd (if isListTailVariableUnif hd then mkOutputMode else mode) tree in
     List.iter (retrieve ~pos:(pos+1) ~add_result mode path) sub_tries
   else begin
-    (* Here the constructor can be Constant, Primitive, Variable, Other, ListHead, ListEnd *)
+    (* Here the constructor can be Constant, Primitive, Variable, Other, ListHead, ListNil, ListEnd *)
     begin
       if get_all_children hd mode then 
         (* we take all the children in the map *)
         on_all_children ~pos:(pos+1) ~add_result mode path map
-      else if isInput mode && isVariable hd then () (* no search has to be done in the map *)
+      else if isInput mode && isVariable hd then
+        try retrieve ~pos:(pos+1) ~add_result mode path (Ptmap.find mkUvarConstant map)
+        with Not_found -> ()
       else
-          (* we have a Constant, Primitive, ListHead or ListHead and look for the key in the map *)
+          (* we have a Constant, Primitive, ListHead, ListNil or ListEnd and look for the key in the map *)
           try retrieve ~pos:(pos+1) ~add_result mode path (Ptmap.find hd map)
           with Not_found -> ()
     end;
@@ -372,13 +389,14 @@ and on_all_children ~pos ~add_result mode path map =
         end
   and skip_functor arity = function
     | Trie.Node { other; map } as tree ->
-        Option.iter (retrieve ~pos ~add_result mode path) other;
-        if arity = 0 then retrieve ~pos ~add_result mode path tree
-        else
-          Ptmap.iter (fun k v ->
-              if isListHead k then skip_list 1 (arity - 1) v
-              else skip_functor (arity - 1 + arity_of k) v)
-            map
+      if arity = 0 then retrieve ~pos ~add_result mode path tree
+      else begin
+        Option.iter (skip_functor (arity-1)) other;
+        Ptmap.iter (fun k v ->
+            if isListHead k then skip_list 1 (arity - 1) v
+            else skip_functor (arity - 1 + arity_of k) v)
+          map
+      end
   in
   Ptmap.iter (fun k v ->
     if isListHead k then skip_list 1 0 v
@@ -387,7 +405,7 @@ and on_all_children ~pos ~add_result mode path map =
 
 let empty_dt args_depth : 'a t =
   let max_depths = Array.make (List.length args_depth) 0 in
-  {t = Trie.empty; max_depths; max_size = 0; max_list_length=0}
+  {t = Trie.empty; max_depths; max_size = 0; max_list_length=0; user_upper_bound_depth = args_depth}
 
 let retrieve ~pos ~add_result path index =
   let mode = Path.get path pos in
@@ -421,4 +439,6 @@ module Internal = struct
   let isListTailVariableUnif = isListTailVariableUnif
   let isPathEnd = isPathEnd
 
+  let isUvarConstant = isUvarConstant
+  let isListNil = isListNil
 end
\ No newline at end of file
