@@ -8,6 +8,12 @@ module type Show = sig
   val show : t -> string
 end
 
+module type ShowKey = sig
+  type key
+  val pp_key : Format.formatter -> key -> unit
+  val show_key : key -> string
+end
+
 module type Show1 = sig
   type 'a t
   val pp : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
@@ -120,15 +126,22 @@ val uniqq: 'a list -> 'a list
 val for_all2 : ('a -> 'a -> bool) -> 'a list -> 'a list -> bool
 val for_all23 :  argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> bool
 val for_all3b : ('a -> 'a -> bool -> bool) -> 'a list -> 'a list -> bool list -> bool -> bool
-type arg_mode = Input | Output
-[@@deriving show, ord]
 
-type mode_aux =
-  | Fo of arg_mode
-  | Ho of arg_mode * mode
-and mode = mode_aux list
+module type Mode = sig
+  type t = Input | Output [@@deriving show, ord]
 
-val for_all3b3 : argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> mode -> bool -> bool
+  type ho = Fo of t | Ho of t * ho list
+  and hos = ho list [@@deriving show, ord]
+
+  val get_head : ho -> t
+  val to_ho : t -> ho
+  val show_pretty : t -> string
+  val pretty : Format.formatter -> t -> unit
+end
+
+module Mode : Mode
+
+val for_all3b3 : argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> Mode.ho list -> bool -> bool
 (*uses physical equality and calls anomaly if the element is not in the list*)
 val remove_from_list : 'a -> 'a list -> 'a list
 (* returns Some t where f x = Some t for the first x in the list s.t.
@@ -237,12 +250,13 @@ val anomaly : ?loc:Loc.t -> string -> 'a
 (* If we type check the program, then these are anomalies *)
 val type_error : ?loc:Loc.t -> string -> 'a
 (* A non fatal warning *)
-val warn : ?loc:Loc.t -> string -> unit
+type warning_id = LinearVariable | UndeclaredGlobal | FlexClause | ImplicationPrecedence
+val warn : ?loc:Loc.t -> id:warning_id -> string -> unit
 (* Indirection for standard print functions *)
 val printf : ('a, Format.formatter, unit) format -> 'a
 val eprintf : ('a, Format.formatter, unit) format -> 'a
 
-val set_warn : (?loc:Loc.t -> string -> unit) -> unit
+val set_warn : (?loc:Loc.t -> id:warning_id -> string -> unit) -> unit
 val set_error : (?loc:Loc.t -> string -> 'a) -> unit
 val set_anomaly : (?loc:Loc.t -> string -> 'a) -> unit
 val set_type_error : (?loc:Loc.t -> string -> 'a) -> unit
