@@ -2,6 +2,7 @@ let us = {|
 
 main :- p0, p1, p2, p3, p4, p5, p6, p7, p8, p9.
 
+type p0, p1, p2, p3, p4, p5, p6, p7, p8, p9 prop.
 |}
 
 let ex = {|
@@ -30,13 +31,12 @@ let () =
   let open Sepcomp.Sepcomp_template in
   let elpi = init () in
   let flags = Compile.default_flags in
-  let us = cc ~elpi ~flags 1 us in
-  let ex = cc ~elpi ~flags 2 ex in
-  let p = Compile.assemble ~elpi [us] in
+  let p,_ = cc ~elpi ~flags ~base:(Compile.empty_base ~elpi) 1 us in
+  let _, ex = cc ~elpi ~flags ~base:p 2 ex in
   let exs = list_init 0 50000 (fun _ -> ex) in
   let rec extend i p =
     if i = 0 then p
-    else extend (i-1) (Compile.extend ~base:p exs) in
+    else extend (i-1) (List.fold_left (fun p u -> Compile.extend ~base:p u) p exs) in
   let p = extend 1 p in
   let q = Compile.query p (Parse.goal_from ~elpi ~loc:(Ast.Loc.initial "g") (Lexing.from_string "main")) in
   exec q
\ No newline at end of file
