@@ -226,7 +226,10 @@ let () = declare "list_comma"
 let () = declare "heap_discard"
   ~source_elpi:"heap_discard.elpi"
   ~description:"heapification of _"
-    ~expectation:Test.Success
+  ()
+let () = declare "heap_discard3"
+  ~source_elpi:"heap_discard3.elpi"
+  ~description:"heapification of _"
   ()
 
 let () = declare "accumulate_twice1"
@@ -527,3 +530,16 @@ let () = declare "implbang"
   ~description:"tail cut =!=>"
   ~expectation:Success
   ()
+
+
+let () = declare "dupta"
+  ~source_elpi:"dupta.elpi"
+  ~description:"duplicate type abbreviation"
+  ~expectation:(FailureOutput (Str.regexp "Illegal type"))
+  ()
+
+let () = declare "dupta2"
+  ~source_elpi:"dupta2.elpi"
+  ~description:"duplicate type abbreviation namespace"
+  ~expectation:(FailureOutput (Str.regexp "Illegal type"))
+  ()
